//-----------------------------------------------------------------------------
// Init
//-----------------------------------------------------------------------------

// const requestlib = require("request");
const intercept = require("intercept-stdout");
const vm = require('vm');
const acorn = require('acorn');
const walk = require( 'estree-walker' ).walk;
const _ = require('lodash')
const debug = require('debug')('sandbox:sandbox');


//-----------------------------------------------------------------------------
// Main
//-----------------------------------------------------------------------------
const hijackReturnStatement = function( code ){
  var timeA = Date.now()

  var ast = acorn.parse(code);
  var returnInstances

  walk( ast, {
    enter: function ( node, parent ) {
      if( node.type == 'VariableDeclarator' && node.id.name == "test"  ){
        returnInstances = node.init.body.body
        this.skip()//stop recursive walk when this is found.
      }
    },
  });

  var returnDoc = _.filter( returnInstances, function ( each ){
    return each.type === "ReturnStatement"
  })

  // debug('mainFunction', JSON.stringify( returnDoc, null, 4 ) );
  // debug('return argument', code.slice( returnDoc[0].argument.start, returnDoc[0].argument.end ) );
  var returnArg = code.slice( returnDoc[0].argument.start, returnDoc[0].argument.end )
  newCode = code.slice( 0 ,returnDoc[0].start ) + 'done('+ returnArg + ')' + code.slice( returnDoc[0].end )
  // debug('mainFunction', newCode );


  debug('time:', Date.now() - timeA );
  return newCode
}

const sandbox = function(script, whiteListedLibraries, defaultContext, callback ){


////////////// PREP ARGS FOR VM //////////////////////
  timeA =  Date.now()
  self = this
  
  var sandboxOutput = {}
  var exitCode = 0
  var sandboxLogs = []
  var sandboxError

  if ( !defaultContext ) {
    var defaultContext = {}
  }
    //standard utils
  defaultContext.setTimeout = setTimeout
  defaultContext.console = console
  defaultContext.Buffer = Buffer
  defaultContext.JSON = JSON

  if ( whiteListedLibraries ) {
    for (var i = 0, len = whiteListedLibraries.length; i < len; i++) {
      // import each library with require as necessary.
      defaultContext[ whiteListedLibraries[i].namespace ]   =  require( whiteListedLibraries[i].npm );
    }
  }



  /// TODO: GROSS I don't like this. How can we tell between sync and asyc scripts

  if ( script.indexOf( 'done(') > -1 ){
    var wrappedCode = "(function(done) {" + script + "; })"
  }else{
    var wrappedCode = "(function(done) {" + script + "; done()})"
  }

  try {
    debug('sandbox.js - running');
    //ASYNC
    timeA =  Date.now()

    //intercept logs
    //https://www.npmjs.com/package/intercept-stdout
    interceptedLogs = [];
    //this initializes the intercept and calling it will end the intercept
    var endIntercept = intercept(function(txt) {
        interceptedLogs.push(txt) //+= txt;
    });

    // TODO: NODE 4.5.0 implementation
    var newContext = vm.createContext( defaultContext )

    var sandbox2 = vm.runInContext( wrappedCode, newContext )(function(callbackResult){
       //the sandbox returns a string. Need to parse again
       var sandboxResult

      //for async Done() syntax
      if( callbackResult ){
        sandboxResult = callbackResult
      }else{
        sandboxResult = newContext.RESULT
      }

      sandboxOutput = { status: 'COMPLETE',result: sandboxResult }
      exitCode = 0

      //stop intercepting console logs
      endIntercept();
      sandboxOutput.console = interceptedLogs

      debug('vm interceptedLogs',  interceptedLogs );

      if ( callback && exitCode === 0  ){
        callback(null, sandboxOutput)
      }
       //  exits with error or kill signal
      if ( callback && (exitCode != 0 ) ){
        callback( sandboxOutput , null)
      }
    })

    // if ( process.version.indexOf('v0.10.')> -1 ){
    //   //////////////// NODE 0.10 implementation: /////////////////
    //   //  contextify needs to be run new everytime.
    //
    //   var Contextify = new require('contextify');
    //   try{
    //     Contextify = Contextify( defaultContext )
    //
    //   }catch(e){
    //     debug('sandbox.js - contextify error', e);
    //   }
    //
    //   Contextify.run( wrappedCode )(function( callbackResult ) {
    //     var sandboxResult
    //
    //     //for async Done() syntax
    //     if( callbackResult ){
    //       sandboxResult = callbackResult
    //     }else{
    //       sandboxResult = Contextify.RESULT
    //     }
    //
    //     sandboxOutput = { status: 'COMPLETE',result: sandboxResult }
    //
    //     exitCode = 0
    //     //kills the sandbox after the timeout
    //     Contextify.dispose();
    //
    //     //stop intercepting console logs
    //     endIntercept();
    //     sandboxOutput.console = interceptedLogs
    //
    //     //  console.log("This text is not being captured");
    //     debug('Contextify - interceptedLogs',  interceptedLogs );
    //
    //
    //     if ( callback && exitCode === 0  ){
    //      callback(null, sandboxOutput)
    //     }
    //
    //     //  exits with error or kill signal
    //     if ( callback && (exitCode != 0 ) ){
    //      callback( sandboxOutput , null)
    //     }
    //
    //     debug('Contextify - Async Complete')
    //   });
    //   //////////////////////// END NODE 0.10 implementation ////////////////////////
    //
    // }

  }catch (e){
   // send message back to parent sandbox.js  with handled  exceptions from the sandbox
    //endIntercept may not have ended
    endIntercept()
    sandboxOutput = { status: 'ERROR', result: e.name +" : " +  e.message, console:  interceptedLogs}
    debug('sandbox.js - error: ', sandboxOutput);
    // process.send('sandbox.js - error: ', sandboxOutput);

    exitCode = 1
    if ( callback && (exitCode != 0 ) ){
      callback( sandboxOutput , null)
    }
  }
}


//-----------------------------------------------------------------------------
// EXPORTS
//-----------------------------------------------------------------------------

module.exports = sandbox;
